"""VSPEC geometry module

This module stores and computes all of the geometry necessary
to simulate a `VSPEC` observation.
"""
import numpy as np
from astropy import units as u
from scipy.optimize import newton
from astropy.table import QTable

import matplotlib.pyplot as plt

class SystemGeometry:
    """System Geometry

    Class to store and calculate information on the geometry of a star-planet-observer system.

    Parameters
    ----------
    inclination : astropy.units.quantity.Quantity [angle], default=0 deg
        The inclination of the system, defined the same as in PSG. Transit is i=90 deg.
    init_stellar_lon : astropy.units.quantity.Quantity [angle], default=0 deg
        The sub-observer longitude at the beginning of observation.
    init_planet_phase : astropy.units.quantity.Quantity [angle], default=0 deg
        The planet phase at beginning of observation.
    stellar_period : astropy.units.quantity.Quantity [time], default=80 day
        The rotational period of the star.
    orbital_period : astropy.units.quantity.Quantity [time], default=11 day
        The orbital period of the planet.
    semimajor_axis : astropy.units.quantity.Quantity [distance], default=0.05 AU
        The semimajor axis of the planet's orbit.
    planetary_rot_period : astropy.units.quantity.Quantity [time], default=11 day
        The rotational period of the planet.
    planetary_init_substellar_lon : astropy.units.quantity.Quantity [angle], default=0 deg
        The sub-stellar longitude of the planet at the beginning of observation.
    stellar_offset_amp : astropy.units.quantity.Quantity [angle], default=0 deg
        The offset between stellar rotation axis and normal to orbital plane.
    stellar_offset_phase : astropy.units.quantity.Quantity [angle], default=0 deg
        The direction of the stellar offset. 0 is defined as facing the observer, and
        right hand direction is positive.
    eccentricity : float, default=0
        The orbital eccentricity of the planet.
    phase_of_periasteron : astropy.units.quantity.Quantity [angle], default=0 deg
        The phase at which the planet reaches periasteron.
    system_distance : astropy.units.quantity.Quantity [distance], default=1.3 pc
        The distance to the system.
    obliquity : astropy.units.quantity.Quantity [angle], default=0 deg
        The planet obliquity magnitude.
    obliquity_direction : astropy.units.quantity.Quantity [angle], default=0 deg
        The true anomaly at which the planet's north pole faces away from the star.


    Attributes
    ----------
    inclination : astropy.units.quantity.Quantity [angle]
        Inclination of the planet's orbit, defined as the angle between the line of
        sight and the orbital plane. Default is 0 deg.
    init_stellar_lon : astropy.units.quantity.Quantity [angle]
        Sub-observer longitude at the beginning of observation. Default is 0 deg.
    init_planet_phase : astropy.units.quantity.Quantity [angle]
        Phase of the planet at the beginning of observation. Default is 0 deg.
    stellar_period : astropy.units.quantity.Quantity [time]
        Rotational period of the star. Default is 80 day.
    orbital_period : astropy.units.quantity.Quantity [time]
        Orbital period of the planet. Default is 11 day.
    semimajor_axis : astropy.units.quantity.Quantity [length]
        Semimajor axis of the planet's orbit. Default is 0.05 AU.
    planetary_rot_period : astropy.units.quantity.Quantity [time]
        Rotational period of the planet. Default is 11 day.
    planetary_init_substellar_lon : astropy.units.quantity.Quantity [angle]
        Sub-stellar longitude of the planet at the beginning of observation. Default is 0 deg.
    alpha : astropy.units.quantity.Quantity [angle]
        Offset between the stellar rotation axis and normal to orbital plane. Default is 0 deg.
    beta : astropy.units.quantity.Quantity [angle]
        Direction of the stellar offset, defined as the angle between the line connecting
        the observer and the system barycenter and the projection of the stellar offset
        vector onto the plane perpendicular to the line of sight. Default is 0 deg.
    eccentricity : float
        Eccentricity of the planet's orbit. Default is 0.
    phase_of_periasteron : astropy.units.quantity.Quantity [angle]
        Phase at which the planet reaches periasteron. Default is 0 deg.
    system_distance : astropy.units.quantity.Quantity [length]
        Distance to the system. Default is 1.3 pc.
    obliquity : astropy.units.quantity.Quantity [angle]
        Obliquity of the planet, defined as the angle between the planet's equator
        and its orbital plane. Default is 0 deg.
    obliquity_direction : astropy.units.quantity.Quantity [angle]
        The true anomaly at which the planet's North pole faces away from the star.
        Default is 0 deg.
    init_time_since_periasteron : astropy.units.quantity.Quantity [angle]
        Time since periasteron at the beginning of observation. Default is the time
        since periasteron at `init_planet_phase`.
    init_true_anomaly : astropy.units.quantity.Quantity [angle]
        True anomaly at the beginning of observation, computed from `init_time_since_periasteron`.
    """

    def __init__(self, inclination=0*u.deg,
                 init_stellar_lon=0*u.deg,
                 init_planet_phase=0*u.deg,
                 stellar_period=80*u.day,
                 orbital_period=11*u.day,
                 semimajor_axis=0.05*u.AU,
                 planetary_rot_period=11*u.day,
                 planetary_init_substellar_lon=0*u.deg,
                 stellar_offset_amp=0*u.deg,
                 stellar_offset_phase=0*u.deg,
                 eccentricity=0,
                 phase_of_periasteron=0*u.deg,
                 system_distance=1.3*u.pc,
                 obliquity=0*u.deg,
                 obliquity_direction=0*u.deg):
        self.inclination = inclination
        self.init_stellar_lon = init_stellar_lon
        self.init_planet_phase = init_planet_phase
        self.stellar_period = stellar_period
        self.orbital_period = orbital_period
        self.semimajor_axis = semimajor_axis
        self.planetary_rot_period = planetary_rot_period
        self.planetary_init_substellar_lon = planetary_init_substellar_lon
        self.alpha = stellar_offset_amp
        self.beta = stellar_offset_phase
        self.eccentricity = eccentricity
        self.phase_of_periasteron = phase_of_periasteron
        self.system_distance = system_distance
        self.obliquity = obliquity
        self.obliquity_direction = obliquity_direction
        self.init_time_since_periasteron = self.get_time_since_periasteron(
            self.init_planet_phase)
        self.init_true_anomaly = self.true_anomaly(
            self.init_time_since_periasteron)

    def sub_obs(self, time):
        """
        Calculate the point on the stellar surface that is facing the observer.

        Parameters
        ----------
        time astropy.units.Quantity [time]
            The time elapsed since the observation began.

        Returns
        -------
        dict
            Coordinates of the sub-observer point in the form {'lat':lat,'lon':lon}
        """
        lon = self.init_stellar_lon - time * 360*u.deg/self.stellar_period
        lat = -1*(90*u.deg - self.inclination) + self.alpha*np.cos(self.beta)
        return {'lat': lat, 'lon': lon % (360*u.deg)}

    def mean_motion(self):
        """
        Get the mean motion of the planet's orbit.

        Returns
        -------
        astropy.units.Quantity [angular frequency]
            The mean motion of the orbit.
        """
        return 360*u.deg / self.orbital_period

    def mean_anomaly(self, time):
        """
        Get the mean anomaly of the orbit at a given time.

        Parameters
        ----------
        time : astropy.units.Quantity [time]
            The time elapsed since periasteron.

        Returns
        -------
        astropy.units.Quantity [angle]
            The mean anomaly.
        """
        return (time * self.mean_motion()) % (360*u.deg)

    def eccentric_anomaly(self, time):
        """
        Calculate the eccentric anomaly of the system
        at a given time

        Parameters
        ----------
        time : astropy.units.Quantity [time]
            The time elapsed since periasteron.

        Returns
        -------
        astropy.units.Quantity [angle]
            The eccentric anomaly.
        """
        mean_anom = self.mean_anomaly(time)

        def func(eccentric_anom: float):
            return mean_anom.to_value(u.rad) - (eccentric_anom*u.deg).to_value(u.rad) + self.eccentricity*np.sin((eccentric_anom*u.deg).to_value(u.rad))
        eccentric_anom = newton(func, x0=30)*u.deg
        return eccentric_anom

    def true_anomaly(self, time):
        """
        Calculate the true anomaly.

        Parameters
        ----------
        time : astropy.units.Quantity [time]
            The time elapsed since periasteron.

        Returns
        -------
        astropy.units.Quantity [angle]
            The true anomaly.
        """
        eccentric_anomaly = self.eccentric_anomaly(time) % (360*u.deg)
        if np.abs((180*u.deg - eccentric_anomaly)/u.deg) < 0.1:
            return eccentric_anomaly
        elif np.abs((0*u.deg - eccentric_anomaly)/u.deg) < 0.1:
            return eccentric_anomaly
        elif np.abs((360*u.deg - eccentric_anomaly)/u.deg) < 0.1:
            return eccentric_anomaly
        else:
            nu0 = 180*u.deg - (180*u.deg-eccentric_anomaly)*0.1

            def func(nu):
                eq = (1-self.eccentricity) * np.tan(nu*u.deg/2)**2 - \
                    (1+self.eccentricity) * np.tan(eccentric_anomaly/2)**2
                return eq
            nu = newton(func, x0=nu0.to_value(u.deg))*u.deg
            return nu

    def phase(self, time):
        """
        Calculate the phase at a given time.
        
        Phase is defined using the PSG convention.
        Transit occurs at phase=180 degrees.

        Parameters
        ----------
        time : astropy.units.Quantity [time]
            The time elapsed since periasteron.

        Returns
        -------
        astropy.units.Quantity [angle]
            The phase.
        """
        return (self.true_anomaly(time) + self.phase_of_periasteron) % (360*u.deg)

    def sub_planet(self, time, phase=None):
        """
        Get the coordinates of the sub-planet point on the star.

        Parameters
        ----------
        time : astropy.units.Quantity [time]
            The time elapsed since periasteron.
        phase : astropy.units.Quantity [angle], default=None
            The current phase, if known. Otherwise phase is calculated based on `time`.
            It is best practice to specify `phase` when possible to avoid using the Newtonian
            solver to calculate it.

        Returns
        -------
        dict
            Coordinates of the sub-planet point in the form {'lat':lat,'lon':lon}
        """
        sub_obs = self.sub_obs(time)
        if phase is None:
            phase = self.phase(time)
        lon = (sub_obs['lon'] + phase + 180*u.deg) % (360*u.deg)
        lat = -1*self.alpha * np.cos(self.beta + phase)
        return {'lat': lat, 'lon': lon}

    def get_time_since_periasteron(self, phase):
        """
        Calculate the time since the last periasteron for a given phase.
        
        This calculation is costly, so it is prefered that the user avoid
        calling this function more than necessary. The output is stored as
        `self.time_since_periasteron` upon initialization, so it can be
        accessed easily without recalculating (so long as no attributes change).

        Parameters
        ----------
        phase : astropy.units.Quantity [angle]
            The current phase of the planet.

        Returns
        -------
        astropy.units.Quantity [time]
            The time elapsed since periasteron.
        """
        true_anomaly = phase - self.phase_of_periasteron
        true_anomaly = true_anomaly % (360*u.deg)
        guess = true_anomaly/360/u.deg * self.orbital_period

        def func(guess):
            val = (self.true_anomaly(guess*u.day) -
                   true_anomaly).to(u.rad).value
            return val
        time = newton(func, (guess/u.day).to(u.Unit(''))) * u.day
        return time.to(u.day)

    def get_substellar_lon_at_periasteron(self) -> u.Quantity:
        """
        Compute the sub-stellar longitude at the previous periasteron
        given the rotation period, orbital period, and initial
        substellar longitude

        Returns
        -------
        astropy.units.Quantity [angle]
            The sub-stellar longitude at periasteron.
        """
        init_time_since_periasteron = self.get_time_since_periasteron(
            self.init_planet_phase)
        init_deg_rotated = 360*u.deg * init_time_since_periasteron/self.planetary_rot_period
        return self.planetary_init_substellar_lon + init_deg_rotated - self.init_true_anomaly

    def get_substellar_lon(self, time_since_periasteron) -> u.quantity.Quantity:
        """
        Calculate the sub-stellar longitude at a particular time since periasteron.

        Parameters
        ----------
        time_since_periasteron : astropy.units.Quantity [time]
            The time elapsed since periasteron.

        Returns
        -------
        astropy.units.Quantity [angle]
            The current sub-stellar longitude of the planet.

        Raises
        ------
        NotImplementedError
            If `self.obliquity` is not 0 deg.
        """
        substellar_lon_at_periasteron = self.get_substellar_lon_at_periasteron()
        deg_rotated = 360*u.deg * time_since_periasteron/self.planetary_rot_period
        true_anom = self.true_anomaly(time_since_periasteron)
        # how long since summer in N
        north_season = (true_anom - self.obliquity_direction) % (360*u.deg)
        if self.obliquity != 0*u.deg:
            raise NotImplementedError(
                'VSPEC does not currently support non-zero obliquity')
        lon = substellar_lon_at_periasteron + true_anom - \
            deg_rotated - self.obliquity*np.cos(north_season)
        # lon = self.planetary_init_substellar_lon - dphase + rotated
        # lon = self.planetary_init_substellar_lon + dphase - rotated
        return lon % (360.0*u.deg)

    def get_substellar_lat(self, phase: u.Quantity) -> u.Quantity:
        """
        Calculate the sub-stellar latitude of the planet at a particular phase.

        Parameters
        ----------
        phase : astropy.units.Quantity [angle]
            The current phase of the planet.

        Returns
        -------
        astropy.units.Quantity [angle]
            The current sub-stellar longitude of the planet.

        Raises
        ------
        NotImplementedError
            If `self.obliquity` is not 0 deg.
        """
        true_anomaly = phase - self.phase_of_periasteron
        north_season = (true_anomaly - self.obliquity_direction) % (360*u.deg)
        if self.obliquity != 0*u.deg:
            raise NotImplementedError(
                'VSPEC does not currently support non-zero obliquity')
        lat = 0*u.deg + self.obliquity*np.cos(north_season)
        return lat

    def get_pl_sub_obs_lon(self, time_since_periasteron: u.quantity.Quantity, phase: u.quantity.Quantity) -> u.quantity.Quantity:
        """
        Compute the sub-observer longitude of the planet.

        Parameters
        ----------
        time_since_periasteron : astropy.units.Quantity [time]
            The time elapsed since periasteron.
        phase : astropy.units.Quantity [angle]
            The current phase of the planet.

        Returns
        -------
        astropy.units.Quantity [angle]
            The current sub-observer longitude of the planet.
        """
        lon = self.get_substellar_lon(time_since_periasteron) - phase
        return lon

    def get_pl_sub_obs_lat(self, phase: u.Quantity) -> u.Quantity:
        """
        Compute the sub-observer latitude of the planet.

        Parameters
        ----------
        phase : astropy.units.Quantity [angle]
            The current phase of the planet.

        Returns
        -------
        astropy.units.Quantity [angle]
            The current sub-observer latitude of the planet.

        Raises
        ------
        NotImplementedError
            If `self.obliquity` is not 0 deg.
        """
        true_anomaly = phase - self.phase_of_periasteron
        north_season = (true_anomaly - self.obliquity_direction) % (360*u.deg)
        if self.obliquity != 0*u.deg:
            raise NotImplementedError(
                'VSPEC does not currently support non-zero obliquity')
        lat = 0*u.deg - self.obliquity * \
            np.cos(north_season) - (90*u.deg-self.inclination)
        return lat

    def get_radius_coeff(self, phase: u.quantity.Quantity) -> float:
        """
        Compute the orbital radius coefficient that depends on eccentricity and phase.

        Parameters
        ----------
        phase : astropy.units.Quantity [angle]
            The current phase of the planet.

        Returns
        -------
        float
            The orbital radius coefficient.
        """
        true_anomaly = phase - self.phase_of_periasteron
        num = 1 - self.eccentricity**2
        den = 1 + self.eccentricity*np.cos(true_anomaly)
        return (num/den).to_value(u.dimensionless_unscaled)

    def get_observation_plan(self,start_times: u.quantity.Quantity,
                             ) -> QTable:
        """
        Calculate information describing the state of the system
        for a series of observations.

        Parameters
        ----------
        start_times : astropy.units.Quantity [time]
            The time of each observation.

        Returns
        -------
        QTable
            The geometry of each observation.
        """
        start_times = start_times.to(u.s)
        
        phases = []
        sub_obs_lats = []
        sub_obs_lons = []
        sub_planet_lats = []
        sub_planet_lons = []
        sub_stellar_lons = []
        sub_stellar_lats = []
        pl_sub_obs_lons = []
        pl_sub_obs_lats = []
        orbit_radii = []
        u_angle = u.deg
        for time in start_times + self.init_time_since_periasteron:
            phase = self.phase(time).to_value(u.deg)  # % (360*u.deg)
            phases.append(phase)
            sub_obs = self.sub_obs(time)
            sub_obs_lats.append(sub_obs['lat'].to_value(u_angle))
            sub_obs_lons.append(sub_obs['lon'].to_value(u_angle))
            sub_planet = self.sub_planet(time, phase=phase*u_angle)
            sub_planet_lats.append(sub_planet['lat'].to_value(u_angle))
            sub_planet_lons.append(sub_planet['lon'].to_value(u_angle))
            sub_stellar_lon = self.get_substellar_lon(time)
            sub_stellar_lat = self.get_substellar_lat(phase*u_angle)
            sub_stellar_lons.append(sub_stellar_lon.to_value(u_angle))
            sub_stellar_lats.append(sub_stellar_lat.to_value(u_angle))
            pl_sub_obs_lon = self.get_pl_sub_obs_lon(time, phase*u_angle)
            pl_sub_obs_lat = self.get_pl_sub_obs_lat(phase*u_angle)
            pl_sub_obs_lons.append(pl_sub_obs_lon.to_value(u_angle))
            pl_sub_obs_lats.append(pl_sub_obs_lat.to_value(u_angle))
            orbit_rad = self.get_radius_coeff(phase*u_angle)
            orbit_radii.append(orbit_rad)
        return QTable(data={'time': start_times,
                'phase': phases*u_angle,
                'sub_obs_lat': sub_obs_lats*u_angle,
                'sub_obs_lon': sub_obs_lons*u_angle,
                'sub_planet_lat': sub_planet_lats*u_angle,
                'sub_planet_lon': sub_planet_lons*u_angle,
                'sub_stellar_lon': sub_stellar_lons*u_angle,
                'sub_stellar_lat': sub_stellar_lats*u_angle,
                'planet_sub_obs_lon': pl_sub_obs_lons*u_angle,
                'planet_sub_obs_lat': pl_sub_obs_lats*u_angle,
                'orbit_radius': orbit_radii})


    def get_system_visual(self,phase:u.Quantity,ax=None) -> plt.Axes:
        """
        Create a graphical representation of the host+planet geometry.

        Parameters
        ----------
        phase : astropy.units.Quantity
            The current phase of the planet.
        ax : matplotlib.axes.Axes, optional
            The axis to draw the figure to, by default None.

        Returns
        -------
        matplotlib.axes.Axes
            The axis with the graphics drawn.
        """
        if ax is None:
            ax = plt.gca()
        ax.set_aspect('equal', adjustable='box')
        ax.scatter(0, 0, c='xkcd:tangerine', s=150)
        theta = np.linspace(0, 360, 180, endpoint=False)*u.deg
        r_dist = (1-self.eccentricity**2)/(1+self.eccentricity*np.cos(theta -
                                                                      self.phase_of_periasteron-90*u.deg))
        curr_theta = phase + 90*u.deg
        x_dist = self.semimajor_axis * np.cos(theta)*r_dist
        y_dist = -1*self.semimajor_axis * \
            np.sin(theta)*r_dist*np.cos(self.inclination)

        current_r = (1-self.eccentricity**2)/(1+self.eccentricity*np.cos(curr_theta -
                                                                         self.phase_of_periasteron - 90*u.deg))
        current_x_dist = self.semimajor_axis * np.cos(curr_theta)*current_r
        current_y_dist = -1*self.semimajor_axis * \
            np.sin(curr_theta)*current_r*np.cos(self.inclination)
        behind = np.sin(theta) >= 0
        x_angle = np.arctan(x_dist/self.system_distance).to(u.mas)
        y_angle = np.arctan(y_dist/self.system_distance).to(u.mas)
        plotlim = np.arctan(
            self.semimajor_axis/self.system_distance).to(u.mas).value * (1+self.eccentricity)*1.05
        current_x_angle = np.arctan(
            current_x_dist/self.system_distance).to(u.mas)
        current_y_angle = np.arctan(
            current_y_dist/self.system_distance).to(u.mas)
        z_order_mapper = {True: -99, False: 100}
        ax.plot(x_angle[behind], y_angle[behind],
                           zorder=-100, c='C0', alpha=1, ls=(0, (2, 2)))
        ax.plot(
            x_angle[~behind], y_angle[~behind], zorder=99, c='C0')
        ax.scatter(current_x_angle, current_y_angle,
                              zorder=z_order_mapper[np.sin(curr_theta) >= 0], c='k')

        ax.set_xlim(-plotlim, plotlim)
        ax.set_ylim(-plotlim, plotlim)
        ax.set_xlabel('sep (mas)')
        ax.set_ylabel('sep (mas)')
        return ax

    def get_planet_visual(self,phase:u.Quantity,ax=None):
        """
        Draw a visualization of the planet's geometry from the view
        of the observer.

        Parameters
        ----------
        phase : astropy.units.Quantity
            The current phase.
        ax : matplotlib.axes.Axes, optional
            The Axes to draw on, by default None

        Returns
        -------
        matplotlib.axes.Axes
            The drawn figure.
        """
        import cartopy.crs as ccrs
        from cartopy.geodesic import Geodesic
        if ax is None:
            ax = plt.gca()
        time_since_periasteron = self.get_time_since_periasteron(phase)
        substellar_lon = self.get_substellar_lon(time_since_periasteron)
        substellar_lat = self.get_substellar_lat(phase)
        
        ax.stock_img()
        lats = np.linspace(-90, 90, 181)*u.deg
        lons = np.linspace(0, 360, 181)*u.deg
        latgrid, longrid = np.meshgrid(lats, lons)
        latgrid = latgrid
        longrid = longrid
        cos_c = (np.sin(substellar_lat) * np.sin(latgrid)
                 + np.cos(substellar_lat) * np.cos(latgrid)
                 * np.cos(substellar_lon-longrid))
        dayside = (cos_c > 0).astype('int')

        # sub_stellar point
        rad_meters = 1e6
        circle_points = Geodesic().circle(lon=substellar_lon.to_value(u.deg), lat=substellar_lat.to_value(u.deg),
                                          radius=rad_meters, n_samples=200, endpoint=False)
        circ_lons = np.array(circle_points[:, 0])
        circ_lats = np.array(circle_points[:, 1])
        hemi_1 = circ_lons > 0
        ax.plot(circ_lons[hemi_1], circ_lats[hemi_1],
                            c='r', transform=ccrs.PlateCarree())
        ax.plot(circ_lons[~hemi_1], circ_lats[~hemi_1],
                            c='r', transform=ccrs.PlateCarree())

        # sub_stellar hemisphere
        rad_meters = (1*u.R_earth).to_value(u.m) * np.pi * 0.5 * 0.99
        circle_points = Geodesic().circle(lon=substellar_lon.to_value(u.deg), lat=substellar_lat.to_value(u.deg),
                                          radius=rad_meters, n_samples=200, endpoint=False)
        circ_lons = np.array(circle_points[:, 0])
        circ_lats = np.array(circle_points[:, 1])
        hemi_1 = circ_lons > 0
        ax.plot(circ_lons[hemi_1], circ_lats[hemi_1],
                            c='r', transform=ccrs.PlateCarree())
        ax.plot(circ_lons[~hemi_1], circ_lats[~hemi_1],
                            c='r', transform=ccrs.PlateCarree())

        ax.plot(
            lons, lons*0,
            transform=ccrs.PlateCarree(),
            c='k'
        )
        ax.plot(
            lons, lons*0+85*u.deg,
            transform=ccrs.PlateCarree(),
            c='C0'
        )
        ax.plot(
            lons, lons*0-85*u.deg,
            transform=ccrs.PlateCarree(),
            c='C1'
        )
        lats = np.linspace(-90, 90)
        ax.plot(lats*0, lats, transform=ccrs.PlateCarree(), c='C2')
        return ax
        
            


    def plot(self, phase: u.Quantity):
        """
        Create a plot of the geometry at a particular phase.

        Parameters
        ----------
        phase : astropy.units.Quantity [angle]
            The current phase of the planet.

        Returns
        -------
        matplotlib.figure.Figure
            A figure containing the plot.
        """
        # import cartopy. This way it is an optional dependencey
        import cartopy.crs as ccrs

        fig = plt.figure()
        axes = {}
        axes['orbit'] = fig.add_subplot(1, 2, 1)
        
        self.get_system_visual(phase,axes['orbit'])
        time_since_periasteron = self.get_time_since_periasteron(phase)
        subobs_lon = self.get_pl_sub_obs_lon(time_since_periasteron, phase)
        subobs_lat = self.get_pl_sub_obs_lat(phase)
        proj = ccrs.Orthographic(
            central_longitude=subobs_lon, central_latitude=subobs_lat)
        axes['planet'] = fig.add_subplot(1, 2, 2, projection=proj)
        self.get_planet_visual(phase,axes['planet'])
        
        return fig



"""VSPEC analysis module

This module is designed to allow a user to easily handle
`VSPEC` outputs.
"""

from typing import Dict, Union, Tuple
from copy import deepcopy
from pathlib import Path
import warnings
import numpy as np
from astropy import units as u
from astropy.io import fits
from datetime import datetime
import json
from astropy.table import QTable
from pypsg import PyLyr

from VSPEC.config import N_ZFILL, MOLEC_DATA_PATH


class PhaseAnalyzer:
    """Class to store and analyze `VSPEC` phase curves

    Read all the data produced from a phase `VSPEC` curve simulation.
    This class also includes some basic-but-powerfull analysis methods meant to be
    quickly used to create nice figures.

    Parameters
    ----------
    path : pathlib.Path or str
        The path to the directory storing all the final output.
        This is usually ``.vspec/.../AllModelSpectraValues/``
    fluxunit : astropy.units.Unit, default=u.Unit('W m-2 um-1')
        Standard unit to use with flux values. This way they are safely
        converted between ``Quantity`` and ``float``

    Examples
    --------
    >>> model = ObservationModel.from_yaml('test.yaml')
    ...
    >>> data = PhaseAnalyzer(model.directories['all_model'])
    >>> plt.plot(data.time, data.lightcurve('total', 0))
    """

    def __init__(self, path, fluxunit=u.Unit('W m-2 um-1')):
        if not isinstance(path, Path):
            path = Path(path)
        self._observation_data: QTable = QTable.read(path / 'observation_info.fits')
        self.n_images: int = len(self._observation_data)
        """
        The total number of epochs simulated.
        
        :type: int
        """
        self.time = self._observation_data['time']
        self.phase = self._observation_data['phase']
        self.unique_phase = deepcopy(self.phase)
        for i in range(len(self.unique_phase) - 1):
            while self.unique_phase[i] > self.unique_phase[i+1]:
                self.unique_phase[i+1] += 360*u.deg
        star = []
        # star_facing_planet = pd.DataFrame() # not super interesting
        reflected = []
        thermal = []
        total = []
        noise = []
        for i in range(self.n_images):
            filename = path / f'phase{str(i).zfill(N_ZFILL)}.fits'
            spectra: QTable = QTable.read(filename)
            if i == 0:  # only do once
                # wavelength
                col = 'wavelength'
                self.wavelength = spectra[col]
            # star
            col = 'star'
            star.append((spectra[col]).to_value(fluxunit))
            # reflected
            col = 'reflected'
            reflected.append((spectra[col]).to_value(fluxunit))
            # reflected
            col = 'planet_thermal'
            thermal.append((spectra[col]).to_value(fluxunit))
            # total
            col = 'total'
            total.append((spectra[col]).to_value(fluxunit))
            # noise
            col = 'noise'
            noise.append((spectra[col]).to_value(fluxunit))
        self.star = np.asarray(star).T * fluxunit
        self.reflected = np.asarray(reflected).T * fluxunit
        self.thermal = np.asarray(thermal).T * fluxunit
        self.total = np.asarray(total).T * fluxunit
        self.noise = np.asarray(noise).T * fluxunit

        try:
            first_lyr:QTable = PyLyr.from_fits(path / f'layer{str(0).zfill(N_ZFILL)}.fits').prof
            colnames = first_lyr.colnames
            vartables: Dict[str, QTable] = {}
            for name in colnames:
                tab = QTable()
                vartables[name] = tab
            
            for i in range(self.n_images):
                filename = path / f'layer{str(i).zfill(N_ZFILL)}.fits'
                dat:QTable = PyLyr.from_fits(filename).prof
                for name in colnames:
                    val = dat[name].value
                    unit = dat[name].unit
                    colname=f'col{i}'
                    vartables[name].add_column(val*unit, name=colname)
            
            self.layers = vartables
        except FileNotFoundError:
            warnings.warn(
                'No Layer info, maybe globes or molecular signatures are off', RuntimeWarning)
            self.layers = fits.HDUList([])

    def _get_mean_molecular_mass(self):
        """
        Get the mean molecular mass
        """
        with open(MOLEC_DATA_PATH, 'rt',encoding='UTF-8') as file:
            molec_data = json.loads(file.read())
        shape = self.get_layer('Alt').shape
        mean_molec_mass = np.zeros(shape=shape)*u.g/u.mol
        for mol, dat in molec_data.items():
            mass = dat['mass']
            try:
                data = self.get_layer(mol)
                mean_molec_mass += data*mass*u.g/u.mol
            except KeyError:
                pass
        return mean_molec_mass
        

    def get_layer(self, var: str) -> u.Quantity:
        """
        Get data from layer variable.

        Access the `self.layers` attribute and return the result as
        a `astropy.units.Quantity` object for a single variable.

        Parameters
        ----------
        var : str
            The name of the variable to access

        Returns
        -------
        astropy.units.Quantity
            They layering data of the requested variable

        Raises
        ------
        KeyError
            If `self` does not have any image data or if `var` is not recognized
        """
        if len(self.layers) == 0:
            raise KeyError('`self.layers` does not contain any data')
        if var == 'MEAN_MASS':
            return self._get_mean_molecular_mass()
        tab: QTable = self.layers[var]
        cols = tab.colnames
        unit = tab[cols[0]].unit
        return np.array(
            [tab[col].to_value(unit) for col in cols],
        )*unit

    def lightcurve(
        self,
        source: str,
        pixel: Union[int, Tuple[int, int]],
        normalize: Union[str,int]='none',
        noise: Union[bool,float,int]=False
    ):
        """
        Produce a lightcurve

        Return the lightcurve of `source` of the wavelengths described by `pixel`

        Parameters
        ----------
        source : str
            Which data array to access.
        pixel : int or 2-tuple of int
            Pixel(s) of spectral axis to use when building lightcurve.
        normalize : str or int, default='none'
            Normalization scheme. If integer, pixel of time axis to
            normalize the lightcurve to. Otherwise it is a keyword
            to describe the normalization process: ``'none'`` or ``'max'``
        noise : bool or float or int, default=False
            Should gaussian noise be added? If float, scale gaussian noise
            by this parameter.

        Returns
        -------
        astropy.units.Quantity
            Lightcurve of the desired source in the desired bandpass

        Raises
        ------
        ValueError
            If `noise` is not `bool`, `float`, or `int`
        ValueError
            If `normalize` is not recognized or `True`

        Warns
        -----
        RuntimeWarning
            If `normalize` is `False`

        """
        if isinstance(pixel, tuple):
            pixel = slice(*pixel)
        flux:np.ndarray = getattr(self, source)[pixel, :]
        if isinstance(noise, bool):
            if noise:
                flux = flux + \
                    np.random.normal(
                        scale=self.noise.value[pixel, :])*self.noise.unit
        elif isinstance(noise, float) or isinstance(noise, int):
            flux = flux + noise * \
                np.random.normal(
                    scale=self.noise.value[pixel, :])*self.noise.unit
        else:
            raise ValueError('noise parameter must be bool, float, or int')
        if flux.ndim > 1:
            flux = flux.mean(axis=0)
        if isinstance(normalize, int):
            flux = (flux/flux[normalize]).to_value(u.dimensionless_unscaled)
        elif isinstance(normalize, str):
            if normalize == 'max':
                flux = (flux/flux.max()).to_value(u.dimensionless_unscaled)
            elif normalize == 'none':
                pass
            else:
                raise ValueError(f'Unknown normalization scheme: {normalize}')
        elif isinstance(normalize, bool):
            if normalize:
                raise ValueError(
                    '`normalize=True` is ambiguous. Please specify a valid scheme')
            else:
                message = 'Setting `normalize=False` can be dangerous as `True` is an ambigous value. '
                message += 'Please use `normalize="none"` instead.'
                warnings.warn(message, RuntimeWarning)
        else:
            raise ValueError(f'Unknown normalization parameter: {normalize}')
        return flux

    def spectrum(
        self,
        source: str,
        images: Union[int, Tuple[int, int]],
        noise: Union[bool,float,int]=False
    ):
        """
        Get a 1D spectrum

        Return the spectrum of a specified source at a single epoch
        or average over multiple epochs.

        Parameters
        ----------
        source : str
            Which data array to access. If `'noise'` is specified, use propagation
            of error formula to calculate theoretical noise of spectrum.
        images : int or 2-tuple of int
            Pixel(s) of time axis to use when building spectrum.
        noise : bool or float or int
            Should gaussian noise be added? If float, scale gaussian noise
            by this parameter.

        Returns
        -------
        astropy.units.Quantity
            Spectrum of the desired source over the desired epoch(s)

        Raises
        ------
        ValueError
            If `noise` is not `bool`, `float`, or `int`
        """
        if isinstance(images, tuple):
            images = slice(*images)
        if source == 'noise':
            flux = self.noise[:, images]**2
            try:
                n_images = flux.shape[1]
                flux = flux.sum(axis=1)/n_images**2
            except IndexError:
                pass
            return np.sqrt(flux)
        else:
            flux = getattr(self, source)[:, images]
            if isinstance(noise, bool):
                if noise:
                    flux = flux + \
                        np.random.normal(
                            scale=self.noise.value[:, images])*self.noise.unit
            elif isinstance(noise, float) or isinstance(noise, int):
                flux = flux + noise * \
                    np.random.normal(
                        scale=self.noise.value[:, images])*self.noise.unit
            else:
                raise ValueError('noise parameter must be bool, float, or int')
            try:
                flux = flux.mean(axis=1)
            except IndexError:  # images is int
                pass
            return flux

    @property
    def fits(self) -> fits.HDUList:
        """
        Covert to an `astropy.io.fits.HDUList` object

        Returns
        -------
        astropy.io.fits.HDUList
            Data converted to the `.fits` format
        """
        primary = fits.PrimaryHDU()
        primary.header['CREATED'] = datetime.now().strftime('%Y%m%d-%H%M%S%Z')
        primary.header['N_images'] = self.n_images

        cols = []
        for col in self._observation_data.colnames:
            cols.append(fits.Column(
                name=col, array=self._observation_data[col].values, format='D'))
        obs_tab = fits.BinTableHDU.from_columns(cols)
        obs_tab.name = 'OBS'

        time = fits.Column(name='time', array=self.time.value, format='D')
        phase = fits.Column(name='phase', array=self.phase.value, format='D')
        unique_phase = fits.Column(
            name='unique_phase', array=self.unique_phase.value, format='D')
        tab1 = fits.BinTableHDU.from_columns([time, phase, unique_phase])
        tab1.header['U_TIME'] = str(self.time.unit)
        tab1.header['U_PHASE'] = str(self.phase.unit)
        tab1.header['U_UPHASE'] = str(self.unique_phase.unit)
        tab1.name = 'PHASE'
        wavelength = fits.Column(
            name='wavelength', array=self.wavelength.value, format='D')
        tab2 = fits.BinTableHDU.from_columns([wavelength])
        tab2.header['U_WAVE'] = str(self.wavelength.unit)
        tab2.name = 'WAVELENGTH'

        total = fits.ImageHDU(self.total.value)
        total.header['U_FLUX'] = str(self.total.unit)
        total.name = 'TOTAL'
        star = fits.ImageHDU(self.star.value)
        star.header['U_FLUX'] = str(self.star.unit)
        star.name = 'STAR'
        reflected = fits.ImageHDU(self.reflected.value)
        reflected.header['U_FLUX'] = str(self.reflected.unit)
        reflected.name = 'REFLECTED'
        thermal = fits.ImageHDU(self.thermal.value)
        thermal.header['U_FLUX'] = str(self.thermal.unit)
        thermal.name = 'THERMAL'
        noise = fits.ImageHDU(self.noise.value)
        noise.header['U_FLUX'] = str(self.noise.unit)
        noise.name = 'NOISE'

        hdul = fits.HDUList([primary, obs_tab, tab1, tab2,
                            total, star, reflected, thermal, noise])
        hdul = hdul + self.layers
        return fits.HDUList(hdul)

    def write_fits(self, filename: str) -> None:
        """
        Save ``PhaseAnalyzer`` object as a `.fits` file.

        Parameters
        ----------
        filename : str
        """
        hdul = self.fits
        hdul.writeto(filename,overwrite=True)
    def to_twocolumn(self,index:Union[int,Tuple[int,int]],outfile:str,fmt:str='ppm',wl:str ='um'):
        """
        Write data to a two column file that can be used in a retrival.
        
        Parameters
        ----------
        index : int or 2-tuple of int
            The index of the epochs to consider. Passed to ``self.spectrum``.
        outfile : str
            The name of the output file
        fmt : str
            The format of the output file. Either ``'ppm'`` or ``'flambda'``.
        wl : str
            The wavelength unit of the output file. Passed to ``astropy.units.Unit``.
        
        Returns
        -------
        None
        
        Raises
        ------
        ValueError
            If `fmt` is not ``'ppm'`` or ``'flambda'``.
        """
        if fmt == 'ppm':
            flux = (self.spectrum('thermal',index,False)+self.spectrum('reflected',index,False))/self.spectrum('total',index,False) * 1e6
            noise = self.spectrum('noise',index,False)/self.spectrum('total',index,False) * 1e6
            flux_unit = u.dimensionless_unscaled
        elif fmt == 'flambda':
            flux = self.spectrum('thermal',index,False)+self.spectrum('reflected',index,False)
            noise = self.spectrum('noise',index,False)
            flux_unit = u.Unit('W m-2 um-1')
        else:
            raise ValueError(f'Unknown format "{fmt}"')
        wl_unit = u.Unit(wl)
        wl = self.wavelength.to_value(wl_unit)
        flux = flux.to_value(flux_unit)
        noise = noise.to_value(flux_unit)
        with open(outfile,'wt',encoding='ascii') as file:
            for w,f,n in zip(wl,flux,noise):
                file.write(f'{w:<10.4f}{f:<14.4e}{n:<14.4e}\n')




"""
File Helpers
"""
from pathlib import Path
import warnings


def check_and_build_dir(path: Path) -> None:
    """
    Check and build directory.

    Check if a path exists, if not, create it.

    Parameters
    ----------
    path : pathlib.Path or str
        Path to check and create
    """
    if isinstance(path, str):
        path = Path(str)
    if path.exists():
        pass
    else:
        path.mkdir(parents=True)


def get_filename(N: int, n_zfill: int, ext: str):
    """
    Get the filename that a PSG output is written to.
    This function unifies all of the filename handling to a single piece of code.

    Parameters
    ----------
    N : int
        The 'phase number' of the file. Represents the iteration of the simulation.
    n_zfill : int
        The ``zfill()`` convention to use for zero-padding the phase number.
    ext : str
        The file extension, such as ``'.rad'`` for rad files.

    Returns
    -------
    str
        The filename

    Warns
    -----
    RuntimeWarning
        If `N` has more digits than `n_zfill` can accommodate.
    """
    if N > (10**(n_zfill)-1):
        warnings.warn(
            f'zfill of {n_zfill} not high enough for phase {N}', RuntimeWarning)
    return f'phase{str(N).zfill(n_zfill)}.{ext}'



from astropy import units as u
from numpy import isclose as np_isclose


def isclose(quant1: u.Quantity, quant2: u.Quantity, tol: u.Quantity) -> bool:
    """
    Check if two quantities are close

    Use `numpy.isclose` on two quantity objects.
    This function safely casts them to floats first.

    Parameters
    ----------
    param1 : astropy.units.Quantity
        First object for comparison
    param2 : astropy.units.Quantity
        Second object for comparison
    tol : astropy.units.Quantity
        Error tolerance between `param1` and `param2`

    Returns
    -------
    bool
        Whether `param1` and `param2` are within `tol`

    Raises
    ------
    astropy.units.UnitConversionError
        If `param1` and `param2` have different physical types.


    Examples
    --------
    >>> from astropy import units as u
    >>> import numpy as np

    >>> values1 = np.array([1.0, 2.0, 3.0]) * u.m
    >>> values2 = np.array([1.01, 2.02, 3.03]) * u.m
    >>> tol = 0.05 * u.m
    >>> isclose(values1, values2, tol)
    array([ True,  True, False])

    >>> temperatures1 = np.array([25.0, 30.0, 35.0]) * u.K
    >>> temperatures2 = np.array([25.5, 30.2, 34.8]) * u.K
    >>> tol = 0.3 * u.K
    >>> isclose(temperatures1, temperatures2, tol)
    array([ True,  True,  True])

    """
    unit = tol.unit
    return np_isclose(quant1.to_value(unit), quant2.to_value(unit), atol=tol.to_value(unit))



"""
Misc helpers
"""

from astropy import units as u
import numpy as np


def get_planet_indicies(
    planet_times: u.Quantity,
    tindex: u.Quantity
    ) -> tuple[int, int]:
    """
    Get the indices of the planet spectra to interpolate over.

    This helper function enables interpolation of planet spectra by determining
    the appropriate indices in the `planet_times` array. By running PSG once for
    multiple "integrations" and interpolating between the spectra, computational
    efficiency is improved.


    Parameters
    ----------
    planet_times : astropy.units.Quantity
        The times (cast to since periasteron) at which the planet spectrum was taken.
    tindex : astropy.units.Quantity
        The epoch of the current observation. The goal is to place this between
        two elements of `planet_times`

    Returns
    -------
    int
        The index of `planet_times` before `tindex`
    int
        The index of `planet_times` after `tindex`

    Raises
    ------
    ValueError
        If multiple elements of 'planet_times' are equal to 'tindex'.
    """
    after = planet_times > tindex
    equal = planet_times == tindex
    if equal.sum() == 1:
        N1 = np.argwhere(equal)[0][0]
        N2 = np.argwhere(equal)[0][0]
    elif equal.sum() > 1:
        raise ValueError('There must be a duplicate time')
    elif equal.sum() == 0:
        N2 = np.argwhere(after)[0][0]
        N1 = N2 - 1
    return N1, N2


"""
Planetary and System Parameters Module
"""

from astropy import units as u
from VSPEC.params.base import BaseParameters
from VSPEC.config import planet_distance_unit, period_unit, planet_radius_unit


class GravityParameters(BaseParameters):
    """
    Class representing gravity parameters.

    Parameters
    ----------
    mode : str
        The mode of the gravity parameter. Valid options are 'g', 'rho', and 'kg'.
    value : astropy.units.Quantity
        The value of the gravity parameter.

    Notes
    -----
    - The available modes and their corresponding units are:
        - 'g': meters per second squared (:math:`{\\rm m~s}^{-2}`)
        - 'rho': grams per cubic centimeter (:math:`{\\rm g~cm}^{-3}`)
        - 'kg': kilograms (kg)

    Examples
    --------
    >>> GravityParameters('g',1000*u.cm/u.s).value
    10.0

    A ``GravityParameters`` object can also be created from a dictionary.

    >>> d = {'mode':'kg','value':1*u.M_earth}
    >>> grav = GravityParameters.from_dict(d)
    >>> grav.mode
    'kg'
    >>> grav.value
    5.972e+24

    It can also write a dictionary to be read by PSG.

    >>> grav.to_psg()
    {'OBJECT-GRAVITY': 5.9720e+24, 'OBJECT-GRAVITY-UNIT': 'kg'}

    """

    _psg_units = {
        'g': u.Unit('m s-2'),
        'rho': u.Unit('g cm-3'),
        'kg': u.kg
    }

    def __init__(
        self,
        mode: str,
        value: u.Quantity
    ):
        self.mode = mode
        self._value = value

    @property
    def value(self):
        """
        The value of the gravity parameter converted to the appropriate unit based on the mode.

        Returns
        -------
        float
            The value of the gravity parameter.
        """

        return self._value.to(self._psg_units[self.mode])

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            mode=str(d['mode']),
            value=u.Quantity(d['value'])
        )

    @classmethod
    def from_dict(cls, d: dict, *args):
        """
        Construct a `GravityParameters` object from a dictionary.

        Parameters
        ----------
        d : dict
            The dictionary to use to construct the class.

        Returns
        -------
        GravityParameters
            The constructed class instance.

        Notes
        -----
        This constructor assumes ``d`` contains the keys ``'mode'`` and ``'value'``
        """
        return super().from_dict(d, *args)


class PlanetParameters(BaseParameters):
    """
    Class representing planet parameters.

    Parameters
    ----------
    name : str
        The name of the planet. Any characters are allowed.
        If you find that PSG is changing the name internally
        please submit an issue.
    radius : astropy.units.Quantity
        The radius of the planet.
    gravity : GravityParameters
        The gravity parameters of the planet.
    semimajor_axis : astropy.units.Quantity
        The semi-major axis of the planet's orbit.
    orbit_period : astropy.units.Quantity
        The period of the planet's orbit.
    rotation_period : astropy.units.Quantity
        The rotation period of the planet.
    eccentricity : float
        The eccentricity of the planet's orbit
    obliquity : astropy.units.Quantity
        The obliquity (tilt) of the planet.
    obliquity_direction : astropy.units.Quantity
        The direction of the planet's obliquity. The true anomaly
        at which the planet's north pole faces away from the star.
    init_phase : astropy.units.Quantity
        The initial phase of the planet.
    init_substellar_lon : astropy.units.Quantity
        The initial substellar longitude of the planet.

    Attributes
    ----------
    name : str
        The name of the planet.
    radius : astropy.units.Quantity
        The radius of the planet.
    gravity : GravityParameters
        The gravity parameters of the planet.
    semimajor_axis : astropy.units.Quantity
        The semi-major axis of the planet's orbit.
    orbit_period : astropy.units.Quantity
        The period of the planet's orbit.
    rotation_period : astropy.units.Quantity
        The rotation period of the planet.
    eccentricity : float
        The eccentricity of the planet's orbit
    obliquity : astropy.units.Quantity
        The obliquity (tilt) of the planet.
    obliquity_direction : astropy.units.Quantity
        The direction of the planet's obliquity. The true anomaly
        at which the planet's north pole faces away from the star.
    init_phase : astropy.units.Quantity
        The initial phase of the planet.
    init_substellar_lon : astropy.units.Quantity
        The initial substellar longitude of the planet.

    Notes
    -----
    - The `proxcenb` class method initializes parameters specific to Proxima Centauri b.
    - The `std` class method initializes parameters for a standard exoplanet.

    """

    def __init__(
        self,
        name: str,
        radius: u.Quantity,
        gravity: GravityParameters,
        semimajor_axis: u.Quantity,
        orbit_period: u.Quantity,
        rotation_period: u.Quantity,
        eccentricity: float,
        obliquity: u.Quantity,
        obliquity_direction: u.Quantity,
        init_phase: u.Quantity,
        init_substellar_lon: u.Quantity
    ):
        self.name = name
        self.radius = radius
        self.gravity = gravity
        self.semimajor_axis = semimajor_axis
        self.orbit_period = orbit_period
        self.rotation_period = rotation_period
        self.eccentricity = eccentricity
        self.obliquity = obliquity
        self.obliquity_direction = obliquity_direction
        self.init_phase = init_phase
        self.init_substellar_lon = init_substellar_lon

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            name=str(d['name']),
            radius=u.Quantity(d['radius']),
            gravity=GravityParameters.from_dict(d['gravity']),
            semimajor_axis=u.Quantity(d['semimajor_axis']),
            orbit_period=u.Quantity(d['orbit_period']),
            rotation_period=u.Quantity(d['rotation_period']),
            eccentricity=float(d['eccentricity']),
            obliquity=u.Quantity(d['obliquity']),
            obliquity_direction=u.Quantity(d['obliquity_direction']),
            init_phase=u.Quantity(d['init_phase']),
            init_substellar_lon=u.Quantity(d['init_substellar_lon'])
        )

    @classmethod
    def proxcenb(cls, init_phase: u.Quantity, init_substellar_lon: u.Quantity):
        """
        Proxima Centauri b :cite:p:`2022A&A...658A.115F`

        Parameters
        ----------
        init_phase : astropy.units.Quantity
            The initial phase of the planet.
        init_substellar_lon : astropy.units.Quantity
            The initial subsolar longitude of the planet.

        Returns
        -------
        PlanetParameters
            The planet parameters for Proxima Centauri b.


        """
        return cls(
            name='Prox Cen b',
            radius=1.03*u.M_earth,  # Earth density
            gravity=GravityParameters('kg', 1.07*u.M_earth),
            semimajor_axis=0.04856*u.AU,
            orbit_period=11.1868*u.day,
            rotation_period=11.1868*u.day,
            eccentricity=0.,
            obliquity=0*u.deg,
            obliquity_direction=0*u.deg,
            init_phase=init_phase,
            init_substellar_lon=init_substellar_lon
        )

    @classmethod
    def std(cls, init_phase: u.Quantity, init_substellar_lon: u.Quantity):
        """
        The default VSPEC planet.

        Parameters
        ----------
        init_phase : astropy.units.Quantity
            The initial phase of the planet.
        init_substellar_lon : astropy.units.Quantity
            The initial subsolar longitude of the planet.

        Returns
        -------
        PlanetParameters
            The planet parameters for the default VSPEC exoplanet.

        """
        return cls(
            name='Exoplanet',
            radius=1.*u.R_earth,
            gravity=GravityParameters('kg', 1.0*u.M_earth),
            semimajor_axis=0.05*u.AU,
            orbit_period=10*u.day,
            rotation_period=10*u.day,
            eccentricity=0.,
            obliquity=0*u.deg,
            obliquity_direction=0*u.deg,
            init_phase=init_phase,
            init_substellar_lon=init_substellar_lon
        )


class SystemParameters(BaseParameters):
    """
    Class representing system parameters.

    Parameters
    ----------
    distance : astropy.units.Quantity
        The distance to the system.
    inclination : astropy.units.Quantity
        The inclination angle of the system. Transit occurs at 90 degrees.
    phase_of_periasteron : astropy.units.Quantity
        The phase (as seen from the observer) of the planet when it reaches periasteron.

    Attributes
    ----------
    distance : astropy.units.Quantity
        The distance to the system.
    inclination : astropy.units.Quantity
        The inclination angle of the system. Transit occurs at 90 degrees.
    phase_of_periasteron : astropy.units.Quantity
        The phase (as seen from the observer) of the planet when it reaches periasteron.

    """

    def __init__(
        self,
        distance: u.Quantity,
        inclination: u.Quantity,
        phase_of_periasteron: u.Quantity
    ):
        self.distance = distance
        self.inclination = inclination
        self.phase_of_periasteron = phase_of_periasteron

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            distance=u.Quantity(d['distance']),
            inclination=u.Quantity(d['inclination']),
            phase_of_periasteron=u.Quantity(d['phase_of_periasteron']),
        )



"""
Observation parameters
"""
from typing import Union
from astropy import units as u
import yaml

from VSPEC.config import flux_unit as default_flux_unit, PRESET_PATH
from VSPEC.params.base import BaseParameters, PSGtable, parse_table


class ObservationParameters(BaseParameters):
    """
    Class storing parameters for observations.

    Parameters
    ----------
    observation_time : astropy.units.Quantity
        The total duration of the observation.
    integration_time : astropy.units.Quantity
        The integration time of each epoch of observation.


    Attributes
    ----------
    observation_time : astropy.units.Quantity
        The total duration of the observation.
    integration_time : astropy.units.Quantity
        The integration time of each epoch of observation.

    Raises
    ------
    ValueError
        If the integration time is longer than the total observation time.

    """

    def __init__(
        self,
        observation_time: u.Quantity,
        integration_time: u.Quantity,
    ):
        self.observation_time = observation_time
        self.integration_time = integration_time
        self._validate()

    def _validate(self):
        if self.integration_time > self.observation_time:
            raise ValueError(
                'Length of integrations cannot be longer than the total observation.')

    @property
    def total_images(self) -> int:
        """
        Total number of images based on the observation time and integration time.

        Returns
        -------
        int
            The total number of images.

        """

        return int(round((self.observation_time/self.integration_time).to_value(u.dimensionless_unscaled)))

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            observation_time=u.Quantity(d['observation_time']),
            integration_time=u.Quantity(d['integration_time']),
        )


class BandpassParameters(BaseParameters):
    """
    Class to store bandpass parameters for observations.

    Parameters
    ----------
    wl_blue : astropy.units.Quantity
        The shorter wavelength limit of the bandpass.
    wl_red : astropy.units.Quantity
        The longer wavelength limit of the bandpass.
    resolving_power : int
        The resolving power of the observation.
    wavelength_unit : astropy.units.Unit
        The unit of wavelength used for the bandpass parameters.
    flux_unit : astropy.units.Unit
        The unit of flux used for the bandpass parameters.

    Attributes
    ----------
    wl_blue : astropy.units.Quantity
        The shorter wavelength limit of the bandpass.
    wl_red : astropy.units.Quantity
        The longer wavelength limit of the bandpass.
    resolving_power : int
        The resolving power of the observation.
    wavelength_unit : astropy.units.Unit
        The unit of wavelength used for the bandpass parameters.
    flux_unit : astropy.units.Unit
        The unit of flux used for the bandpass parameters.
    """

    psg_rad_mapper = {u.Unit('W m-2 um-1'): 'Wm2um'}

    def __init__(
        self,
        wl_blue: u.Quantity,
        wl_red: u.Quantity,
        resolving_power: int,
        wavelength_unit: u.Unit,
        flux_unit: u.Unit
    ):
        self.wl_blue = wl_blue
        self.wl_red = wl_red
        self.resolving_power = resolving_power
        self.wavelength_unit = wavelength_unit
        self.flux_unit = flux_unit

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            wl_blue=u.Quantity(d['wl_blue']),
            wl_red=u.Quantity(d['wl_red']),
            resolving_power=int(d['resolving_power']),
            wavelength_unit=u.Unit(d['wavelength_unit']),
            flux_unit=u.Unit(d['flux_unit'])
        )

    @classmethod
    def mirecle(cls):
        """
        MIRECLE setup [1].
        Returns a `BandpassParameters` instance initialized with the MIRECLE setup.

        Returns
        -------
        BandpassParameters:
            `BandpassParameters` instance with the MIRECLE setup.

        References
        ----------
        [1] :cite:t:`2022AJ....164..176M`
        """
        return cls(
            1*u.um,
            18*u.um,
            50,
            u.um,
            default_flux_unit
        )

    @classmethod
    def miri_lrs(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['bandpass']['miri-lrs'])

    @classmethod
    def niriss_soss(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['bandpass']['niriss-soss'])


class ccdParameters(BaseParameters):
    """
    Class to store CCD parameters for observations.

    Parameters
    ----------
    pixel_sampling : int
        The pixel sampling of the CCD.
    read_noise : astropy.units.Quantity
        The read noise of the CCD in electrons.
    dark_current : astropy.units.Quantity
        The dark current of the CCD in electrons/second.
    throughput : float
        The throughput of the CCD.
    emissivity : float
        The emissivity of the CCD.
    temperature : astropy.units.Quantity
        The temperature of the CCD.


    Attributes
    ----------
    pixel_sampling : int
        The pixel sampling of the CCD.
    read_noise : astropy.units.Quantity
        The read noise of the CCD in electrons.
    dark_current : astropy.units.Quantity
        The dark current of the CCD in electrons/second.
    throughput : float
        The throughput of the CCD.
    emissivity : float
        The emissivity of the CCD.
    temperature : astropy.units.Quantity
        The temperature of the CCD.
    """

    def __init__(
        self,
        pixel_sampling: int,
        read_noise: u.Quantity,
        dark_current: u.Quantity,
        throughput: float,
        emissivity: float,
        temperature: u.Quantity
    ):
        self.pixel_sampling = pixel_sampling
        self.read_noise = read_noise
        self.dark_current = dark_current
        self.throughput = throughput
        self.emissivity = emissivity
        self.temperature = temperature

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            pixel_sampling=parse_table(d['pixel_sampling'], int),
            read_noise=parse_table(d['read_noise'], u.Quantity),
            dark_current=parse_table(d['dark_current'], u.Quantity),
            throughput=parse_table(d['throughput'], float),
            emissivity=float(d['emissivity']),
            temperature=u.Quantity(d['temperature'])
        )

    @classmethod
    def mirecle(cls):
        """
        MIRECLE setup [1].
        Returns a CCDParameters instance initialized with the MIRECLE setup.

        Returns
        -------
        CCDParameters:
            CCDParameters instance for the MIRECLE setup.

        References
        ----------
        [1] :cite:t:`2022AJ....164..176M`
        """

        return cls(
            pixel_sampling=64,
            read_noise=6 * u.electron,
            dark_current=100 * u.electron / u.s,
            throughput=0.5,
            emissivity=0.1,
            temperature=35*u.K
        )

    @classmethod
    def miri_lrs(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['ccd']['miri-lrs'])

    @classmethod
    def niriss_soss(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['ccd']['niriss-soss'])


class DetectorParameters(BaseParameters):
    """
    Class to store detector parameters for observations.

    Parameters
    ----------
    beam_width : astropy.units.Quantity
        The beam width of the detector.
    integration_time : astropy.units.Quantity
        The integration time of the detector.
    ccd : ccdParameters
        The CCD parameters for the detector.

    Attributes
    ----------
    beam_width : astropy.units.Quantity
        The beam width of the detector.
    integration_time : astropy.units.Quantity
        The integration time of the detector.
    ccd : ccdParameters
        The CCD parameters for the detector.
    """

    def __init__(
        self,
        beam_width: u.Quantity,
        integration_time: u.Quantity,
        ccd: ccdParameters
    ):
        self.beam_width = beam_width
        self.integration_time = integration_time
        self.ccd = ccd

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            beam_width=u.Quantity(d['beam_width']),
            integration_time=u.Quantity(d['integration_time']),
            ccd=ccdParameters.from_dict(d['ccd']),
        )

    @classmethod
    def miri_lrs(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['detector']['miri-lrs'])

    @classmethod
    def niriss_soss(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['detector']['niriss-soss'])

    @classmethod
    def mirecle(cls):
        """
        MIRECLE setup [1].
        Returns a DetectorParameters instance initialized with the MIRECLE setup.

        Returns
        -------
        DetectorParameters:
            DetectorParameters instance for the MIRECLE setup.

        References
        ----------
        [1] :cite:t:`2022AJ....164..176M`
        """
        return cls(
            beam_width=5*u.arcsec,
            integration_time=0.5*u.s,
            ccd=ccdParameters.mirecle()
        )


class TelescopeParameters(BaseParameters):
    """
    Base class for telescope Parameters.

    Parameters
    ----------
    aperture : astropy.units.Quantity
        The aperture size of the telescope.
    mode : str
        The mode of the telescope. Valid values are 'single' or 'coronagraph'.
    zodi : float
        The level of the zodiacal background. Acceptable values range from
        1.0 (Ecliptic pole/minimum) to 100.0 (Close to ecliptic/Sun).

    Other Parameters
    ----------------
    kwargs
        Additional parameters specific to the telescope.

    Attributes
    ----------
    aperture : astropy.units.Quantity
        The aperture size of the telescope.
    mode : str
        The mode of the telescope.
    zodi : float
        The level of the zodiacal background.
    """

    _mode_translator = {
        'single': 'SINGLE',
        'coronagraph': 'CORONA',
    }

    def __init__(
        self,
        aperture: u.Quantity,
        mode: str,
        zodi: float,
        contrast: float | PSGtable = None,
        iwa: float | PSGtable = None,
    ):
        self.aperture = aperture
        self.mode = mode
        self.zodi = zodi
        self.contrast = contrast
        self.iwa = iwa


class SingleDishParameters(TelescopeParameters):
    """
    Parameters for a single dish telescope.

    Parameters
    ----------
    aperture : astropy.units.Quantity
        The aperture size of the telescope.
    zodi : float
        The level of the zodiacal background. Acceptable values range from
        1.0 (Ecliptic pole/minimum) to 100.0 (Close to ecliptic/Sun).
    """

    def __init__(self, aperture: u.Quantity, zodi: float):
        super().__init__(aperture, 'single', zodi)

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            aperture=u.Quantity(d['aperture']),
            zodi=float(d['zodi'])
        )

    @classmethod
    def mirecle(cls):
        """
        Create a SingleDishParameters instance with MIRECLE parameters [1].

        Returns
        -------
        SingleDishParameters
            The created SingleDishParameters instance with MIRECLE parameters.

        References
        ----------
        [1] :cite:t:`2022AJ....164..176M`
        """

        return cls(2*u.m, 1.0)

    @classmethod
    def jwst(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['telescope']['jwst']['single'])


class CoronagraphParameters(TelescopeParameters):
    """
    Parameters for a coronagraph telescope.

    Parameters
    ----------
    aperture : astropy.units.Quantity
        The aperture size of the telescope.
    zodi : float
        The level of the zodiacal background. Acceptable values range from
        1.0 (Ecliptic pole/minimum) to 100.0 (Close to ecliptic/Sun).
    exozodi : float
        The level of the exozodiacal background.
    contrast : float
        The contrast level.
    IWA : PSGtable
        The inner working angle of the coronagraph
    """

    def __init__(
        self,
        aperture: u.Quantity,
        zodi: float,
        contrast: float,
        iwa: PSGtable
    ):
        super().__init__(
            aperture,
            'coronagraph',
            zodi,
            contrast=contrast,
            iwa=iwa
        )

    @classmethod
    def _from_dict(cls, d: dict):
        return cls(
            aperture=u.Quantity(d['aperture']),
            zodi=float(d['zodi']),
            contrast=float(d['contrast']),
            iwa=PSGtable.from_dict(d['iwa']['table'])
        )


class InstrumentParameters(BaseParameters):
    """
    Class to store instrument parameters for observations.

    Parameters
    ----------
    telescope : TelescopeParameters
        The telescope parameters for the instrument.
    bandpass : BandpassParameters
        The bandpass parameters for the instrument.
    detector : DetectorParameters
        The detector parameters for the instrument.

    Attributes
    ----------
    telescope : TelescopeParameters
        The telescope parameters for the instrument.
    bandpass : BandpassParameters
        The bandpass parameters for the instrument.
    detector : DetectorParameters
        The detector parameters for the instrument.

    """

    def __init__(
        self,
        telescope: Union[SingleDishParameters, CoronagraphParameters],
        bandpass: BandpassParameters,
        detector: DetectorParameters
    ):
        self.telescope = telescope
        self.bandpass = bandpass
        self.detector = detector

    @classmethod
    def _from_dict(cls, d: dict):
        if 'coronagraph' in d.keys():
            telescope = CoronagraphParameters.from_dict(d['coronagraph'])
        elif 'single' in d.keys():
            telescope = SingleDishParameters.from_dict(d['single'])
        else:
            raise KeyError('Cannot find Telescope Parameters key')
        return cls(
            telescope=telescope,
            bandpass=BandpassParameters.from_dict(d['bandpass']),
            detector=DetectorParameters.from_dict(d['detector'])
        )

    @classmethod
    def from_preset(cls, name):
        match name:
            case 'mirecle':
                return cls.mirecle()
            case 'miri_lrs':
                return cls.miri_lrs()
            case 'niriss_soss':
                return cls.niriss_soss()
            case _:
                raise ValueError('Invalid preset name')

    @classmethod
    def mirecle(cls):
        """
        2m MIRECLE setup [1].
        Returns an InstrumentParameters instance initialized with the 2m MIRECLE setup.

        Returns
        -------
        InstrumentParameters
            `InstrumentParameters` instance for the 2m MIRECLE setup.

        References
        ----------
        [1] :cite:t:`2022AJ....164..176M`
        """
        return cls(
            telescope=SingleDishParameters.mirecle(),
            bandpass=BandpassParameters.mirecle(),
            detector=DetectorParameters.mirecle()
        )

    @classmethod
    def miri_lrs(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['instrument']['miri-lrs'])

    @classmethod
    def niriss_soss(cls):
        path = PRESET_PATH / 'jwst.yaml'
        with open(path, 'r', encoding='UTF-8') as file:
            data = yaml.safe_load(file)
            return cls.from_dict(data['instrument']['niriss-soss'])



"""
Helpers for Teff related calculations
"""

from astropy import units as u
import numpy as np
import warnings

from VSPEC import config


def arrange_teff(minteff: u.Quantity, maxteff: u.Quantity):
    """
    Generate a list of effective temperature (Teff) values with steps of 100 K that fully
    encompass the specified range.

    This function is useful for obtaining a list of Teff values to be used for binning spectra later on.

    Parameters
    ----------
    minteff : astropy.units.Quantity
        The minimum Teff value required.
    maxteff : astropy.units.Quantity
        The maximum Teff value required.

    Returns
    -------
    teffs : list of int
        An array of Teff values, with steps of 100 K.

    Notes
    -----
    - The function calculates the Teff values that fully encompass the specified range by rounding down the minimum value to the nearest multiple of 100 K and rounding up the maximum value to the nearest multiple of 100 K.
    - The `np.arange` function is then used to generate a sequence of Teff values with steps of 100 K, covering the entire range from the rounded-down minimum value to the rounded-up maximum value.

    Examples
    --------
    >>> minteff = 5000 * u.K
    >>> maxteff = 6000 * u.K
    >>> arrange_teff(minteff, maxteff)
    [5000, 5100, 5200, 5300, 5400, 5500, 5600, 5700, 5800, 5900, 6000]
    """

    step = 100*u.K
    if (minteff % step) == 0*u.K:
        low = minteff
    else:
        low = minteff - (minteff % step)
    if (maxteff % step) == 0*u.K:
        high = maxteff
    else:
        high = maxteff - (maxteff % step) + step
    number_of_steps = ((high-low)/step).to_value(u.dimensionless_unscaled)
    number_of_steps = int(round(number_of_steps))
    teffs = low + np.arange(number_of_steps+1)*step
    return np.array([int(round(teff.to_value(u.K))) for teff in teffs],dtype=int)


def get_surrounding_teffs(Teff: u.Quantity):
    """
    Get the effective temperatures (Teffs) of the two spectra to interpolate between
    in order to obtain a spectrum with the target Teff.

    This function is useful for determining the Teffs of the two spectra that surround
    a given target Teff value, which are necessary for performing interpolation to
    obtain a spectrum with the desired Teff.

    Parameters
    ----------
    Teff : astropy.units.Quantity
        The target Teff for the interpolated spectrum.

    Returns
    -------
    low_teff : astropy.units.Quantity
        The Teff of the spectrum below the target Teff.
    high_teff : astropy.units.Quantity
        The Teff of the spectrum above the target Teff.

    Raises
    ------
    ValueError
        If the target Teff is a multiple of 100 K, which would cause problems with ``scipy`` interpolation.

    Notes
    -----
    - The function checks if the target Teff is a multiple of 100 K. If it is, a `ValueError` is raised because this would lead to issues with scipy interpolation.
    - If the target Teff is not a multiple of 100 K, the function determines the Teff of the spectrum
        below the target Teff by rounding down to the nearest multiple of 100 K, and the Teff of the
        spectrum above the target Teff is obtained by adding 100 K to the low Teff.

    Examples
    --------
    >>> Teff = 5500 * u.K
    >>> get_surrounding_teffs(Teff)
    (<Quantity 5500. K>, <Quantity 5600. K>)
    """

    step = 100*u.K
    if (Teff % step) == 0*u.K:
        raise ValueError(
            f'Teff of {Teff} is a multiple of {100*u.K}. This will cause problems with scipy.')
    else:
        low_teff = Teff - (Teff % step)
        high_teff = low_teff + step
    return low_teff, high_teff


def round_teff(teff):
    """
    Round the effective temperature to the nearest integer.
    The goal is to reduce the number of unique effective temperatures
    while not affecting the accuracy of the model.

    Parameters
    ----------
    teff : astropy.units.Quantity
        The temperature to round.

    Returns
    -------
    astropy.units.Quantity
        The rounded temperature.

    Notes
    -----
    This function rounds the given effective temperature to the nearest integer value. It is designed to decrease the number of unique effective temperatures while maintaining the accuracy of the model.

    Examples
    --------
    >>> teff = 1234.56 * u.K
    >>> rounded_teff = round_teff(teff)
    >>> print(rounded_teff)
    1235 K

    >>> teff = 2000.4 * u.K
    >>> rounded_teff = round_teff(teff)
    >>> print(rounded_teff)
    2000 K

    """

    val = teff.value
    unit = teff.unit
    return int(round(val)) * unit


def clip_teff(teff: u.Quantity):
    """
    Clip an effective temperature value to ensure it is within the bounds of
    available models.

    Parameters
    ----------
    teff : astropy.units.Quantity
        The effecitve temperature to clip

    Returns
    -------
    astropy.units.Quantity
        The clipped effective temperature.
    """
    low, high = config.grid_teff_bounds
    if teff > high:
        warnings.warn(
            f'Teff of {teff:.1f} too high, clipped to {high:.1f}', RuntimeWarning)
        return high
    elif teff < low:
        warnings.warn(
            f'Teff of {teff:.1f} too low, clipped to {low:.1f}', RuntimeWarning)
        return low
    else:
        return teff



"""VSPEC module to communicate with the PSG API

This module communucates between `VSPEC` and
and the Planetary Spectrum Generator via the API.
"""

import warnings
from astropy import units as u
import numpy as np

import pypsg

from VSPEC.params.read import InternalParameters

warnings.simplefilter('ignore', category=u.UnitsWarning)

def cfg_to_bytes(config:dict)->bytes:
    """
    Convert a PSG config dictionary into a bytes sequence.

    Parameters
    ----------
    config : dict
        The dictionary containing PSG parameters
    
    Returns
    -------
    bytes
        A bytes object containing the file content.
    
    .. deprecated:: 0.4.0
        Functionality moved to ``pypsg``.
    """
    msg = 'Functionality moved to `pypsg`.'
    warnings.warn(msg, DeprecationWarning)
    s = b''
    for key,value in config.items():
        s += bytes(f'<{key}>{value}\n',encoding='UTF-8')
    return s

def cfg_to_dict(config:str)->dict:
    """
    Convert a PSG config file into a dictionary.
    
    .. deprecated:: 0.4.0
        Functionality moved to ``pypsg``.
    """
    msg = 'Functionality moved to `pypsg`.'
    warnings.warn(msg, DeprecationWarning)
    cfg = {}
    for line in config.split('\n'):
        key,value = line.replace('<','').split('>')
        cfg.update({key:value})
    return cfg

def change_psg_parameters(
    params:InternalParameters,
    phase:u.Quantity,
    orbit_radius_coeff:float,
    sub_stellar_lon:u.Quantity,
    sub_stellar_lat:u.Quantity,
    pl_sub_obs_lon:u.Quantity,
    pl_sub_obs_lat:u.Quantity,
    include_star:bool
    )->pypsg.PyConfig:
    """
    Get the time-dependent PSG parameters

    Parameters
    ----------
    params : VSPEC.params.Parameters
        The parameters of this VSPEC simulation
    phase : astropy.units.Quantity
        The phase of the planet
    orbit_radius_coeff : float
        The planet-star distance normalized to the semimajor axis.
    sub_stellar_lon : astropy.units.Quantity
        The sub-stellar longitude of the planet.
    sub_stellar_lat : astropy.units.Quantity
        The sub-stellar latitude of the planet.
    pl_sub_obs_lon : astropy.units.Quantity
        The sub-observer longitude of the planet.
    pl_sub_obs_lat : astropy.units.Quantity
        The sub-observer latitude of the planet.
    include_star : bool
        If True, include the star in the simulation.
    
    Returns
    -------
    config : dict
        The PSG config in dictionary form.
    """
    target = pypsg.cfg.Target(
        star_type=params.star.psg_star_template if include_star else '-',
        season=phase,
        star_distance=orbit_radius_coeff*params.planet.semimajor_axis,
        solar_longitude=sub_stellar_lon,
        solar_latitude=sub_stellar_lat,
        obs_longitude=pl_sub_obs_lon,
        obs_latitude=pl_sub_obs_lat
    )
    return pypsg.PyConfig(target=target)



def get_reflected(
    cmb_rad: pypsg.PyRad,
    therm_rad: pypsg.PyRad,
    planet_name: str
    ) -> u.Quantity:
    """
    Get reflected spectra.

    Parameters
    ----------
    cmb_rad : PSGrad
        A rad file from the star+planet PSG call.
    therm_rad : PSGrad
        A rad file from the planet-only PSG call.

    Returns
    -------
    astropy.units.Quantity
        The spectrum of the reflected light.

    Raises
    ------
    ValueError
        If the wavelength axes do not match.
    KeyError
        If neither object has a `'Reflected'` data array and at least one
        of them is missing the `planet_name` data array.
    """
    axis_equal = np.all(np.isclose(
        cmb_rad.wl.to_value(u.um),
        therm_rad.wl.to_value(u.um),
        atol=1e-3
    ))
    if not axis_equal:
        raise ValueError('The spectral axes must be equivalent.')
    planet_name = planet_name.replace(' ', '-')

    
    if 'Reflected' in cmb_rad.colnames:
        return cmb_rad['Reflected']
    elif 'Reflected' in therm_rad.colnames:
        return therm_rad['Reflected']
    elif (planet_name in cmb_rad.colnames) and (planet_name in therm_rad.colnames):
        if 'Transit' in cmb_rad.colnames:
            return cmb_rad[planet_name] * 0 # assume there is no refection during transit
        else:
            return cmb_rad[planet_name] - therm_rad[planet_name]
    else:
        raise KeyError(f'Data array {planet_name} not found.')



"""
VSPEC helpers module

This module contains functions that may be shared
throughout the rest of the package.
"""
from VSPEC.helpers.astropy_units import isclose
from VSPEC.helpers.misc import get_planet_indicies
from VSPEC.helpers.teff import arrange_teff, get_surrounding_teffs, round_teff, clip_teff
from VSPEC.helpers.files import check_and_build_dir, get_filename



"""
VSPEC Builtin data.
"""

from pathlib import Path
import requests

BASE_PATH = Path.home() / '.vspec'
DATA_PATH = BASE_PATH / 'data'

WACCM_URL = 'https://zenodo.org/records/10426886/files/vspec_waccm_test.nc?#mode=bytes'

WACCM_PATH = DATA_PATH / 'vspec_waccm_test.nc'
"""
The local path for the WACCM test dataset.

:type: pathlib.Path
"""

def download_waccm_test_data(rewrite=False):
    """
    Download the WACCM test dataset.
    
    Taken from this SO post:
        https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
    
    Parameters
    ----------
    rewrite : bool, optional
        If True, re-download the data. The default is False.
    
    Returns
    -------
    pathlib.Path
        The path to the downloaded data.
    """
    DATA_PATH.mkdir(exist_ok=True)
    WACCM_PATH.parent.mkdir(exist_ok=True)
    if WACCM_PATH.exists() and not rewrite:
        return WACCM_PATH
    else:
        WACCM_PATH.unlink(missing_ok=True)
        with requests.get(WACCM_URL,stream=True,timeout=20) as req:
            req.raise_for_status()
            with WACCM_PATH.open('wb') as f:
                for chunk in req.iter_content(chunk_size=8192):
                    f.write(chunk)
        return WACCM_PATH


"""
The ``VSPEC`` parameters module.

This classes instruct the behavior of the rest
of the ``VSPEC`` package.
"""
from VSPEC.params.base import BaseParameters
from VSPEC.params.read import InternalParameters
from VSPEC.params.read import Header
from VSPEC.params.planet import PlanetParameters
from VSPEC.params.planet import GravityParameters
from VSPEC.params.planet import SystemParameters
from VSPEC.params.gcm import psgParameters
from VSPEC.params.gcm import gcmParameters
from VSPEC.params.observation import ObservationParameters
from VSPEC.params.observation import InstrumentParameters
from VSPEC.params.observation import TelescopeParameters
from VSPEC.params.observation import CoronagraphParameters
from VSPEC.params.observation import SingleDishParameters
from VSPEC.params.observation import DetectorParameters
from VSPEC.params.observation import ccdParameters
from VSPEC.params.observation import BandpassParameters
from VSPEC.params.stellar import StarParameters
from VSPEC.params.stellar import GranulationParameters
from VSPEC.params.stellar import FlareParameters
from VSPEC.params.stellar import FaculaParameters
from VSPEC.params.stellar import SpotParameters
from VSPEC.params.stellar import LimbDarkeningParameters